# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import list2fixture, setup_function, teardown_function

spaces = [ift.GLSpace(5),
          ift.MultiDomain.make({'': ift.RGSpace(5, distances=.789)}),
          (ift.RGSpace(3, distances=.789), ift.UnstructuredDomain(2))]
pmp = pytest.mark.parametrize
field = list2fixture([ift.from_random(sp, 'normal') for sp in spaces] +
                     [ift.from_random(sp, 'normal', dtype=np.complex128) for sp in spaces])

dtype = list2fixture([np.float64,
                     np.complex128])
Nsamp = 2000
np.random.seed(42)


def _to_array(d):
    if isinstance(d, np.ndarray):
        return d
    assert isinstance(d, dict)
    return np.concatenate(list(d.values()))


def _complex2real(sp):
    tup = tuple([d for d in sp])
    rsp = ift.DomainTuple.make((ift.UnstructuredDomain(2),) + tup)
    rl = ift.DomainTupleFieldInserter(rsp, 0, (0,))
    im = ift.DomainTupleFieldInserter(rsp, 0, (1,))
    x = ift.ScalingOperator(sp, 1)
    return rl(x.real)+im(x.imag)


def test_complex2real():
    sp = ift.UnstructuredDomain(3)
    op = _complex2real(ift.makeDomain(sp))
    f = ift.from_random(op.domain, 'normal', dtype=np.complex128)
    assert np.all((f == op.adjoint_times(op(f))).val)
    assert op(f).dtype == np.float64
    f = ift.from_random(op.target, 'normal')
    assert np.all((f == op(op.adjoint_times(f))).val)


def energy_tester(pos, get_noisy_data, energy_initializer, assume_diagonal=False):
    if isinstance(pos, ift.Field):
        if np.issubdtype(pos.dtype, np.complexfloating):
            op = _complex2real(pos.domain)
        else:
            op = ift.ScalingOperator(pos.domain, 1.)
    else:
        ops = []
        for k, dom in pos.domain.items():
            if np.issubdtype(pos[k].dtype, np.complexfloating):
                ops.append(_complex2real(dom).ducktape(k).ducktape_left(k))
            else:
                FA = ift.FieldAdapter(dom, k)
                ops.append(FA.adjoint @ FA)
        realizer = ift.utilities.my_sum(ops)
        from nifty8.operator_spectrum import _DomRemover
        flattener = _DomRemover(realizer.target)
        op = flattener @ realizer
    pos = op(pos)

    domain = pos.domain
    test_vec = ift.from_random(domain, 'normal')
    results = []
    lin = ift.Linearization.make_var(pos)
    for i in range(Nsamp):
        data = get_noisy_data(op.adjoint_times(pos))
        energy = energy_initializer(data) @ op.adjoint
        grad = energy(lin).gradient
        if assume_diagonal:
            results.append(_to_array((grad*grad.conjugate()).val))
        else:
            results.append(_to_array((grad*grad.s_vdot(test_vec)).val))
    if assume_diagonal:
        res = np.mean(np.array(results), axis=0)*_to_array(test_vec.val)
        std = np.std(np.array(results), axis=0)/np.sqrt(Nsamp)*np.abs(_to_array(test_vec.val))
    else:
        res = np.mean(np.array(results), axis=0)
        std = np.std(np.array(results), axis=0)/np.sqrt(Nsamp)
    energy = energy_initializer(data) @ op.adjoint
    lin = ift.Linearization.make_var(pos, want_metric=True)
    res2 = _to_array(energy(lin).metric(test_vec).val)
    np.testing.assert_allclose(res/std, res2/std, atol=5)
    # Test whether one would detect a factor of 2 in the Fisher metric
    for factor in [0.5, 2]:
        with pytest.raises(AssertionError):
            np.testing.assert_allclose(res/std, factor*res2/std, atol=5)


def test_GaussianEnergy(field):
    icov = ift.from_random(field.domain, 'normal')**2
    icov = ift.makeOp(icov, sampling_dtype=field.dtype)
    get_noisy_data = lambda mean: mean + icov.draw_sample(from_inverse=True)
    E_init = lambda data: ift.GaussianEnergy(data=data, inverse_covariance=icov)
    energy_tester(field, get_noisy_data, E_init)


def test_PoissonEnergy(field):
    if not isinstance(field, ift.Field):
        pytest.skip("MultiField Poisson energy  not supported")
    if np.iscomplexobj(field.val):
        pytest.skip("Poisson energy not defined for complex flux")
    get_noisy_data = lambda mean: ift.makeField(mean.domain, np.random.poisson(mean.val))
    # Make rate positive and high enough to avoid bad statistic
    lam = 10*(field**2).clip(0.1, None)
    E_init = lambda data: ift.PoissonianEnergy(data)
    energy_tester(lam, get_noisy_data, E_init)


def test_VariableCovarianceGaussianEnergy(dtype):
    dom = ift.UnstructuredDomain(3)
    res = ift.from_random(dom, 'normal', dtype=dtype)
    ivar = ift.from_random(dom, 'normal')**2+4.
    mf = ift.MultiField.from_dict({'res': res, 'ivar': ivar})
    energy = ift.VariableCovarianceGaussianEnergy(dom, 'res', 'ivar', dtype)
    def get_noisy_data(mean):
        samp = ift.from_random(dom, 'normal', dtype)
        samp = samp/mean['ivar'].sqrt()
        return samp + mean['res']
    def E_init(data):
        adder = ift.Adder(ift.MultiField.from_dict({'res': data}), neg=True)
        return energy.partial_insert(adder)
    energy_tester(mf, get_noisy_data, E_init, assume_diagonal=True)


def normal(dtype, shape):
    return ift.random.Random.normal(dtype, shape)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import itertools

import numpy as np
import pytest
from nifty8 import GLSpace, myassert
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize

# [nlat, nlon, expected]
CONSTRUCTOR_CONFIGS = [[
    2, None, {
        'nlat': 2,
        'nlon': 3,
        'harmonic': False,
        'shape': (6,),
        'size': 6,
    }
], [0, None, {
    'error': ValueError
}]]


def get_dvol_configs():
    wgt = [2.0943951, 2.0943951]
    # for GLSpace(nlat=2, nlon=3)
    dvol_0 = np.array(
        list(
            itertools.chain.from_iterable(
                itertools.repeat(x, 3) for x in wgt)))
    return [
        [1, dvol_0],
    ]


@pmp('attribute', ['nlat', 'nlon'])
def test_property_ret_type(attribute):
    g = GLSpace(2)
    myassert(isinstance(getattr(g, attribute), int))


@pmp('nlat, nlon, expected', CONSTRUCTOR_CONFIGS)
def test_constructor(nlat, nlon, expected):
    g = GLSpace(4)

    if 'error' in expected:
        with assert_raises(expected['error']):
            GLSpace(nlat, nlon)
    else:
        g = GLSpace(nlat, nlon)
        for key, value in expected.items():
            assert_equal(getattr(g, key), value)


@pmp('power, expected', get_dvol_configs())
def test_dvol(power, expected):
    assert_almost_equal(GLSpace(2).dvol, expected)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..minimization.conjugate_gradient import ConjugateGradient
from ..minimization.quadratic_energy import QuadraticEnergy
from .endomorphic_operator import EndomorphicOperator
from .operator import Operator


class SamplingEnabler(EndomorphicOperator):
    """Class which acts as a operator object built of (`likelihood` + `prior`)
    and enables sampling from its inverse even if the operator object
    itself does not support it.


    Parameters
    ----------
    likelihood : :class:`EndomorphicOperator`
        Metric of the likelihood
    prior : :class:`EndomorphicOperator`
        Metric of the prior
    iteration_controller : :class:`IterationController`
        The iteration controller to use for the iterative numerical inversion
        done by a :class:`ConjugateGradient` object.
    approximation : :class:`LinearOperator`, optional
        if not None, this linear operator should be an approximation to the
        operator, which supports the operation modes that the operator doesn't
        have. It is used as a preconditioner during the iterative inversion,
        to accelerate convergence.
    start_from_zero : boolean
        If true, the conjugate gradient algorithm starts from a field filled
        with zeros. Otherwise, it starts from a prior samples. Default is
        False.
    """

    def __init__(self, likelihood, prior, iteration_controller,
                 approximation=None, start_from_zero=False):
        if not isinstance(likelihood, Operator) or not isinstance(prior, Operator):
            raise TypeError
        self._likelihood = likelihood
        self._prior = prior
        self._ic = iteration_controller
        self._approximation = approximation
        self._start_from_zero = bool(start_from_zero)
        self._op = likelihood + prior
        self._domain = self._op.domain
        self._capability = self._op.capability
        self.apply = self._op.apply

    def special_draw_sample(self, from_inverse=False):
        try:
            res = self._op.draw_sample(from_inverse)
            return self._op(res), res
        except NotImplementedError:
            if not from_inverse:
                raise ValueError("from_inverse must be True here")
            if self._start_from_zero:
                b = self._op.draw_sample()
                energy = QuadraticEnergy(0*b, self._op, b)
            else:
                s = self._prior.draw_sample(from_inverse=True)
                nj = self._likelihood.draw_sample()
                b = self._prior(s) + nj
                energy = QuadraticEnergy(s, self._op, b,
                                         _grad=self._likelihood(s) - nj)
            inverter = ConjugateGradient(self._ic)
            if self._approximation is not None:
                energy, convergence = inverter(
                    energy, preconditioner=self._approximation.inverse)
            else:
                energy, convergence = inverter(energy)
            return b, energy.position

    def draw_sample(self, from_inverse=False):
        return self.special_draw_sample(from_inverse=from_inverse)[1]

    def __repr__(self):
        from ..utilities import indent
        return "\n".join((
            "SamplingEnabler:",
            indent("\n".join((
                "Likelihood:", self._likelihood.__repr__(),
                "Prior:", self._prior.__repr__())))))



# Copyright(C) 2022 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause
# Authors: Gordian Edenhofer, Philipp Frank

import abc
from dataclasses import dataclass, field
from functools import partial
from pprint import pformat
from typing import Any, Callable, Iterable, Optional
from warnings import warn

from jax import eval_shape
from jax import numpy as jnp
from jax import random, vmap
from jax.tree_util import (
    register_pytree_node,
    tree_leaves,
    tree_map,
    tree_structure,
    tree_unflatten,
)

from .misc import wrap
from .tree_math import PyTreeString, ShapeWithDtype, random_like, Vector


class Initializer:
    domain = ShapeWithDtype((2,), jnp.uint32)

    def __new__(cls, call_or_struct):
        if isinstance(call_or_struct, Initializer):
            return call_or_struct
        obj = super().__new__(cls)
        obj._call_or_struct = call_or_struct
        obj._target = None  # Used only for caching
        return obj

    def __call__(self, key, *args, **kwargs):
        if not self.stupid:
            struct = tree_structure(self._call_or_struct)
            # Cast the subkeys to the structure of `primals`
            subkeys = tree_unflatten(struct, random.split(key, struct.num_leaves))

            def draw(init, key):
                return init(key, *args, **kwargs)

            return tree_map(draw, self._call_or_struct, subkeys)

        return self._call_or_struct(key, *args, **kwargs)

    @property
    def target(self):
        if self._target is None:
            self._target = eval_shape(self, self.domain)
        return self._target

    @property
    def stupid(self):
        return callable(self._call_or_struct)

    def __or__(self, other):
        other = Initializer(other)
        if not self.stupid and not other.stupid:
            return Initializer(self._call_or_struct | other._call_or_struct)
        # TODO: we can actually do better here and combine the output of both
        # calls in a new call
        return NotImplemented

    def __getitem__(self, key):
        if not self.stupid:
            return Initializer(self._call_or_struct[key])
        raise NotImplementedError("'stupid' initializer not supported")

    def __len__(self):
        if not self.stupid:
            return len(self._call_or_struct)
        return len(self.target)

    def __repr__(self):
        s = "Initializer("
        rep = pformat(self._call_or_struct).replace("\n", "\n\t").strip()
        s += f"\n\t{rep}\n)"
        s = s.replace("\n", "").replace("\t", "") if s.count("\n") <= 2 else s
        return s

    def __str__(self):
        return repr(self)


class ModelMeta(abc.ABCMeta):
    """Register all derived classes as PyTrees in JAX using metaprogramming.

    For any dataclasses.Field property with a metadata-entry named "static",
    we will either hide or expose the property to JAX depending on the value.
    """

    def __new__(mcs, name, bases, dict_, /, **kwargs):
        cls = super().__new__(mcs, name, bases, dict_, **kwargs)
        cls = dataclass(init=False, repr=False, eq=False)(cls)
        IS_STATIC_DEFAULT = True

        def tree_flatten(self):
            static = []
            dynamic = []
            for k, v in self.__dict__.items():
                # Inspired by how equinox registers properties as static in JAX
                fm = self.__dataclass_fields__.get(k)
                fm = fm.metadata if fm is not None else {}
                if fm.get("static", IS_STATIC_DEFAULT) is False:
                    dynamic.append((PyTreeString(k), v))
                else:
                    static.append((k, v))
            return (tuple(dynamic), tuple(static))

        @partial(partial, cls=cls)
        def tree_unflatten(aux, children, *, cls):
            static, dynamic = aux, children
            obj = object.__new__(cls)
            for nm, m in dynamic + static:
                setattr(obj, str(nm), m)  # unwrap any potential `PyTreeSring`s
            return obj

        # Register class and all classes deriving from it
        register_pytree_node(cls, tree_flatten, tree_unflatten)
        return cls


class NoValue:
    pass


class LazyModel(metaclass=ModelMeta):
    _domain: Any = field()
    _target: Any = field()
    _init: Any = field()

    def __init__(self, domain=NoValue, target=NoValue, init=NoValue):
        self._domain = domain
        self._target = target
        self._init = Initializer(init) if init is not NoValue else init

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    @property
    def domain(self):
        if self._domain is NoValue and self._init is not NoValue:
            return eval_shape(self.init, Initializer.domain)
        return self._domain

    @property
    def target(self):
        if self._target is NoValue and self.domain is not NoValue:
            return eval_shape(self.__call__, self.domain)
        return self._target

    @property  # Needs to be a property to enable `model_a.init | model_B.init`
    def init(self) -> Initializer:
        if self._init is NoValue:
            msg = (
                "drawing white parameters"
                ";\nto silence this warning, overload the `init` method"
            )
            warn(msg)
            return Initializer(
                tree_map(lambda p: partial(random_like, primals=p), self.domain)
            )
        return self._init


class Model(LazyModel):
    """Join a callable with a domain, target, and an init method.

    From a domain and a callable, this class automatically derives the target as
    well as instantiate a default initializer if not set explicitly. More
    importantly though, it registers the class as PyTree in JAX using
    metaprogramming. By default all properties are hidden from JAX except those
    marked via `dataclasses.field(metadata=dict(static=False))` as non-static.
    """

    def __init__(
        self,
        call: Optional[Callable] = None,
        *,
        domain=NoValue,
        target=NoValue,
        init=NoValue,
        white_init=False,
    ):
        """Wrap a callable and associate it with a `domain`.

        Parameters
        ----------
        call : callable
            Method acting on objects of type `domain`.
        domain : tree-like structure of ShapeWithDtype, optional
            PyTree of objects with a shape and dtype attribute. Inferred from
            init if not specified.
        target : tree-like structure of ShapeWithDtype, optional
            PyTree of objects with a shape and dtype attribute akin to the
            output of `call`. Inferrred from `call` and `domain` if not set.
        init : callable, optional
            Initialization method taking a PRNG key as first argument and
            creating an object of type `domain`. Inferred from `domain`
            assuming a white normal prior if not set.
        white_init : bool, optional
            If `True`, the domain is set to a white normal prior. Defaults to
            `False`.
        """
        self._call = call
        if init is NoValue and domain is not NoValue and white_init is True:
            init = tree_map(lambda p: partial(random_like, primals=p), domain)
        elif init is NoValue and domain is NoValue:
            raise ValueError("one of `init` or `domain` must be set")

        if domain is NoValue and init is not NoValue:
            domain = eval_shape(init, Initializer.domain)
        if target is NoValue and domain is not NoValue:
            target = eval_shape(self, domain)  # Honor overloaded `__call__`
        super().__init__(domain=domain, init=init, target=target)

    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)

    def __repr__(self):
        s = f"{self.__class__.__name__}("
        if self._call:
            rep = pformat(self._call).replace("\n", "\n\t").strip()
        else:
            rep = f"<bound method {self.__class__.__name__}.__call__>"
        s += f"\n\t{rep}"
        if self._domain is not None:
            rep = pformat(self._domain).replace("\n", "\n\t").strip()
            s += f",\n\tdomain={rep}"
        if self._init is not None:
            rep = pformat(self._init).replace("\n", "\n\t").strip()
            s += f",\n\tinit={rep}"
        s += "\n)"
        s = s.replace("\n", "").replace("\t", "") if s.count("\n") <= 2 else s
        return s

    def __str__(self):
        return repr(self)


class WrappedCall(Model):
    def __init__(
        self,
        call: Callable,
        *,
        name=None,
        shape=(),
        dtype=None,
        white_init=False,
        target=NoValue,
    ):
        """Transforms `call` such that instead of it acting on `input` it
        selects `name` from `input` using `input[name]`.

        Parameters
        ----------
        call : callable
            Callable to wrap.
        name : hashable, optional
            New name of the `input` on which `call` acts.
        shape : tuple or tree-like structure of ShapeWithDtype
            Shape of old `input` on which `call` acts. This can also be an
            arbitrary shape-dtype structure in which case `dtype` is ignored.
            Defaults to a scalar.
        dtype : dtype or tree-like structure of ShapeWithDtype
            If `shape` is a tuple, this is the dtype of the old `input` on
            which `call` acts. This is redundant if `shape` already encodes the
            `dtype`.


        See :class:`Model` for details on the remaining arguments.
        """
        leaves = tree_leaves(shape)
        isswd = all(hasattr(e, "shape") and hasattr(e, "dtype") for e in leaves)
        isswd &= len(leaves) > 0
        domain = ShapeWithDtype(shape, dtype) if not isswd else shape

        if name is not None:
            call = wrap(call, name=name)
            domain = {name: domain}
        super().__init__(call, domain=domain, target=target, white_init=white_init)


def _is_none_or_int(x):
    # Workaround as jax.tree_util.none_leaf_registry is not exposed
    return isinstance(x, int) or (x is None)


def _parse_axes(axes, domain, name=""):
    struct = tree_structure(domain)
    if isinstance(axes, int):
        axes = tree_unflatten(struct, (axes,) * struct.num_leaves)
    else:
        # Shortcut for dict only models
        if isinstance(axes, str):
            axes = (axes,)
        if isinstance(axes, Iterable) and all((isinstance(ii, str) for ii in axes)):
            dom = dict(domain)
            if dom != domain:
                msg = f"{name} must be dict-like if axes are strings"
                raise ValueError(msg)
            axes = {k: (0 if k in axes else None) for k in dom.keys()}

        ax_struct = tree_structure(axes, is_leaf=_is_none_or_int)
        if ax_struct != struct:
            msg = f"{name} structure {struct} does not match axis structure {ax_struct}"
            raise ValueError(msg)
    return axes


class VModel(LazyModel):
    model: LazyModel = field(metadata=dict(static=False))
    in_axes: Any
    out_axes: Any
    axis_size: int

    def __init__(self, model, axis_size, in_axes=0, out_axes=0):
        if not isinstance(model, LazyModel):
            raise ValueError(f"Model {model} of invalid type")
        if model.init.stupid:
            raise ValueError("can only vmap models with a non-'stupid' init")
        self.model = model

        if not isinstance(axis_size, int) and axis_size <= 0:
            raise ValueError(f"invalid axis size {axis_size}")

        self.in_axes = _parse_axes(in_axes, model.domain, "Model domain")
        self.out_axes = _parse_axes(out_axes, model.target, "Model target")

        def _init(key, func, axes):
            ks = random.split(key, axis_size)
            return vmap(func, out_axes=axes)(ks)

        def _parse_init(func, axes):
            if axes is NoValue:
                return func
            return partial(_init, func=func, axes=axes)

        parse_axes = tree_map(
            lambda x: NoValue if x is None else x, self.in_axes, is_leaf=_is_none_or_int
        )
        init = tree_map(_parse_init, self.model.init._call_or_struct, parse_axes)
        super().__init__(init=init)

    def __call__(self, x):
        axs = self.in_axes
        axs_tr = axs.tree if isinstance(axs, Vector) else axs
        x_tr = x.tree if isinstance(x, Vector) else x
        if isinstance(axs_tr, dict) and isinstance(x_tr, dict):
            axs_tr = axs_tr | {k: None for k in x_tr.keys() - axs_tr.keys()}
        axs = Vector(axs_tr) if isinstance(x, Vector) else axs_tr
        return vmap(self.model, (axs,), self.out_axes)(x)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from nifty8 import myassert
from numpy.testing import assert_allclose, assert_raises

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp("scale", [1, 0.5, 0.23, 0.98])
def test_scaled_lh(scale):
    dom = ift.UnstructuredDomain(10)
    d = ift.makeField(dom, (5*ift.from_random(dom)).abs().val.astype(int))
    lh0 = ift.PoissonianEnergy(d) @ ift.Operator.identity_operator(dom).exp()
    lh1 = scale*lh0 + (1-scale)*lh0

    pos = ift.from_random(dom)+1
    pos1 = ift.from_random(dom)

    met0 = lh0.get_metric_at(pos)
    met1 = lh1.get_metric_at(pos)

    ift.extra.assert_allclose(met0(pos1), met1(pos1))
